/* ----
 * ---- file   : mlb.h
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel.
 * ----          Distributed under terms of the MIT LICENSE (MIT).
 * ----
 * ---- Permission is hereby granted, free of charge, to any person obtaining a copy
 * ---- of this software and associated documentation files (the "Software"), to deal
 * ---- in the Software without restriction, including without limitation the rights
 * ---- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * ---- copies of the Software, and to permit persons to whom the Software is
 * ---- furnished to do so, subject to the following conditions:
 * ----
 * ---- The above copyright notice and this permission notice shall be included in
 * ---- all copies or substantial portions of the Software.
 * ----
 * ---- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * ---- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * ---- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * ---- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * ---- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * ---- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * ---- THE SOFTWARE.
 * ----
 * ---- info   : This is part of the "c64_tools" package.
 * ----
 * ---- changed: 10Sep2013, 12Sep2013, 14Sep2013, 15Sep2013, 18Sep2013, 19Sep2013
 * ----
 * ----
 */

#ifndef __MLB_H__
#define __MLB_H__


/* Must at least be 1 */
#define MLB_MAX_COMPONENTS  (16u)


typedef sU16 dsp_component_id_t;
typedef sU16 dsp_component_cmd_t;


/*
 * Utility macro to interpret an unsigned component argument
 *  as a signed value.
 *
 *  (note) this macro is meant for use in dsp_component_exec_fxn functions
 *          (to cast the _arg1 / _arg2 parameters)
 */
#define DSP_ARG_S32(v)  (*(sS32*)&(v))


/*
 * Utility macro to read the _arg1/_arg2 parameters in a dsp_component_exec_fxn.
 *
 *  (note) yep, arg1/2 could be used directly but since there is DSP_ARG_S32, this
 *          macro was added for completeness' sake.
 */
#define DSP_ARG_U32(v)  (*(sU32*)&(v))


/*
 * Utility macro to cast a component id (enum) to an unsigned short.
 */
#define DSP_COMPONENT_ID(v) ((dsp_component_id_t)(v))


/*
 * Utility macro to cast a component command id (enum) to an unsigned short.
 */
#define DSP_COMPONENT_CMD_ID(v) ((dsp_component_cmd_t)(v))


/*
 * Utility macro to define a fastcall cmd (32bit, lower 16bits must be 0)
 */
#define DSP_COMPONENT_FC_CMD_DECLARE(c)  ((c) << 16u)


/*
 * Utility macro to dispatch a fastcall cmd (32bit, lower 16bits must be 0)
 *
 *   (this is mainly done so the compiler can generate a jump table for the switchcase)
 */
#define DSP_COMPONENT_FC_CMD_DISPATCH(c)  ((c) >> 16u)


/*
 * Component "init" callback function pointer type.
 *
 *  (note) When the callback returns S_FALSE, mlb_panic() will be invoked
 *          and the DSP will be stuck in an infinite loop.
 *
 */
typedef sBool (*dsp_component_init_fxn) (void);


/*
 * Component "exec" callback function pointer type.
 *
 *  "exec" functions will be called when a request has been received from the GPP.
 *
 *  The meaning of _cmd depends on the respective component.
 *  Usually this parameter is used as a function selector.
 *
 *  The "ret1" and "ret" output parameters can be used to return data to the GPP side.
 *   (the pointers are always != NULL).
 *
 *  The function should return 0 if the command was executed successfully.
 *
 *  A return value != 0 will signal the mailbox system that the command could
 *  not be executed.
 *
 *  (note) this is the only required component function
 *
 */
typedef sU32 (*dsp_component_exec_fxn) (dsp_component_cmd_t _cmd,
                                        sU32  _arg1, sU32  _arg2,
                                        sU32 *_ret1, sU32 *_ret2
                                        );


/*
 * Component "exit" callback function pointer type.
 *
 */
typedef void (*dsp_component_exit_fxn) (void);


/*
 * Component structure
 *
 *  (note) the component_id is filled out when the component is registered
 *          using mlb_component_register().
 *          It must not be 0.
 *
 */
typedef struct {
   dsp_component_id_t component_id;  /* will be assigned in mlb_component_register() */

   struct {
      dsp_component_init_fxn init;    /* [optional] Called after main() when mlb task is starting up */
      dsp_component_exec_fxn exec;    /* [required] Called when an RPC was received from the GPP */
      dsp_component_exec_fxn exec_fc; /* [optional] Called when a fastcall RPC is detected */
      dsp_component_exit_fxn exit;    /* [optional] not used ATM, should be NULL */
   } fxns;

} dsp_component_t;


/*
 * Initialize mailbox
 *
 * Must be called in main() before registering components
 *
 */
S_EXTERN void mlb_init (void);


/*
 * Register DSP component.
 *
 *  Each component must have a unique id.
 *
 *  The dsp_component_t structure may be allocated on the stack -- it will
 *   be copied by this function.
 *
 *  This function return S_TRUE (1) when the component has been registered successfully.
 *
 */
S_EXTERN sBool mlb_component_register (dsp_component_id_t    _compId,
                                       const dsp_component_t *_comp
                                       );


/*
 * Start mailbox task / create mlb task.
 *
 * Must be called in main() after all components have been registered.
 *
 */
S_EXTERN void mlb_start (void);


/*
 * Write debug value that can be examined on the GPP side by reading 4 bytes from /dev/c64.
 *
 *  (note) idx must be in the range 0..3
 */
S_EXTERN void mlb_debug_usr (sU32 _idx, sU32 _val);


/*
 * This function can be called by DSP components in order to initiate a fastcall sequence.
 */
S_EXTERN void mlb_fastcall_initiate (void);


/*
 * This function should be called when the system (the DSP) becomes unusable.
 *
 *  (note) This function never returns.
 *
 */
S_EXTERN void mlb_panic (void);


#endif /* __MLB_H__ */
